@@ -1,4 +1,4 @@
-;;; derived.el -- allow inheritance of major modes.
+;;; derived.el --- allow inheritance of major modes.
 ;;; (formerly mode-clone.el)
 
 ;; Copyright (C) 1993, 1994 Free Software Foundation, Inc.
@@ -19,8 +19,9 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to
-;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+;; along with GNU Emacs; see the file COPYING.  If not, write to the
+;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+;; Boston, MA 02111-1307, USA.
 
 ;;; Commentary:
 
@@ -73,7 +74,7 @@
 ;;   default, but may be redefined by the user to contain special
 ;;   commands (ie. setting local variables like 'outline-regexp')
 ;;   **NOTE: do not use this option -- it will soon be obsolete.
-;; - run anything assigned to 'hypertext-mode-hooks' (obsolete, but
+;; - run anything assigned to 'hypertext-mode-hook' (obsolete, but
 ;;   supported for the sake of compatibility).
 ;;
 ;; The advantages of this system are threefold.  First, text mode is
@@ -106,8 +107,8 @@
 
 The arguments to this command are as follow:
 
-PARENT:    the name of the command for the parent mode (ie. text-mode).
 CHILD:     the name of the command for the derived mode.
+PARENT:    the name of the command for the parent mode (ie. text-mode).
 NAME:      a string which will appear in the status line (ie. \"Hypertext\")
 DOCSTRING: an optional documentation string--if you do not supply one,
            the function will attempt to invent something useful.
@@ -142,30 +143,31 @@ been generated automatically, with a reference to the keymap."
   (setq docstring (or docstring (derived-mode-make-docstring parent child)))
 
   (` (progn 
-       (derived-mode-init-mode-variables (quote (, child)))
+       (derived-mode-init-mode-variables '(, child))
+       (put '(, child) 'derived-mode-parent '(, parent))
        (defun (, child) ()
 	 (, docstring)
 	 (interactive)
 					; Run the parent.
 	 ((, parent))
 					; Identify special modes.
-	 (if (get (quote (, parent)) 'special)
-	     (put (quote (, child)) 'special t))
+	 (if (get '(, parent) 'special)
+	     (put '(, child) 'special t))
 					; Identify the child mode.
-	 (setq major-mode (quote (, child)))
+	 (setq major-mode '(, child))
 	 (setq mode-name (, name))
 					; Set up maps and tables.
-	 (derived-mode-set-keymap (quote (, child)))
-	 (derived-mode-set-syntax-table (quote (, child)))
-	 (derived-mode-set-abbrev-table (quote (, child)))
+	 (derived-mode-set-keymap '(, child))
+	 (derived-mode-set-syntax-table '(, child))
+	 (derived-mode-set-abbrev-table '(, child))
 					; Splice in the body (if any).
 	 (,@ body)
 ;;;					; Run the setup function, if
 ;;;					; any -- this will soon be
 ;;;					; obsolete.
-;;;	 (derived-mode-run-setup-function (quote (, child)))
-					; Run the hooks, if any.
-	 (derived-mode-run-hooks (quote (, child)))))))
+;;;	 (derived-mode-run-setup-function '(, child))
+					; Run the hook, if any.
+	 (derived-mode-run-hooks '(, child))))))
 
 
 ;; PUBLIC: find the ultimate class of a derived mode.
@@ -185,9 +187,9 @@ Use the `derived-mode-parent' property of the symbol to trace backwards."
   "Construct a setup-function name based on a mode name."
   (intern (concat (symbol-name mode) "-setup")))
 
-(defsubst derived-mode-hooks-name (mode)
-  "Construct a hooks name based on a mode name."
-  (intern (concat (symbol-name mode) "-hooks")))
+(defsubst derived-mode-hook-name (mode)
+  "Construct the mode hook name based on mode name MODE."
+  (intern (concat (symbol-name mode) "-hook")))
 
 (defsubst derived-mode-map-name (mode)
   "Construct a map name based on a mode name."
@@ -221,7 +223,10 @@ the first time the mode is used."
   (if (boundp (derived-mode-syntax-table-name mode))
       t
     (eval (` (defvar (, (derived-mode-syntax-table-name mode))
-	       (make-vector 256 nil)
+	       ;; Make a syntax table which doesn't specify anything
+	       ;; for any char.  Valid data will be merged in by
+	       ;; derived-mode-merge-syntax-tables.
+	       (make-char-table 'syntax-table nil)
 	       (, (format "Syntax table for %s." mode)))))
     (put (derived-mode-syntax-table-name mode) 'derived-mode-unmerged t))
 
@@ -255,8 +260,9 @@ which more-or-less shadow
   (let* ((map-name (derived-mode-map-name mode))
 	 (new-map (eval map-name))
 	 (old-map (current-local-map)))
-    (if (get map-name 'derived-mode-unmerged)
-	(derived-mode-merge-keymaps old-map new-map))
+    (and old-map
+	 (get map-name 'derived-mode-unmerged)
+	 (derived-mode-merge-keymaps old-map new-map))
     (put map-name 'derived-mode-unmerged nil)
     (use-local-map new-map)))
 
@@ -287,9 +293,9 @@ Always merge its parent into it, since the merge is non-destructive."
 ;;;	(funcall fname))))
 
 (defun derived-mode-run-hooks (mode)
-  "Run the hooks if they exist."
+  "Run the mode hook for MODE."
 
-  (let ((hooks-name (derived-mode-hooks-name mode)))
+  (let ((hooks-name (derived-mode-hook-name mode)))
     (if (boundp hooks-name)
 	(run-hooks hooks-name))))
 
@@ -297,19 +303,36 @@ Always merge its parent into it, since the merge is non-destructive."
 
 (defun derived-mode-merge-keymaps (old new)
   "Merge an old keymap into a new one.
-The old keymap is set to be the cdr of the new one, so that there will
+The old keymap is set to be the last cdr of the new one, so that there will
 be automatic inheritance."
+  (let ((tail new))
+    ;; Scan the NEW map for prefix keys.
+    (while (consp tail)
+      (and (consp (car tail))
+	   (let* ((key (vector (car (car tail))))
+		  (subnew (lookup-key new key))
+		  (subold (lookup-key old key)))
+	     ;; If KEY is a prefix key in both OLD and NEW, merge them.
+	     (and (keymapp subnew) (keymapp subold)
+		  (derived-mode-merge-keymaps subold subnew))))
+      (and (vectorp (car tail))
+	   ;; Search a vector of ASCII char bindings for prefix keys.
+	   (let ((i (1- (length (car tail)))))
+	     (while (>= i 0)
+	       (let* ((key (vector i))
+		      (subnew (lookup-key new key))
+		      (subold (lookup-key old key)))
+		 ;; If KEY is a prefix key in both OLD and NEW, merge them.
+		 (and (keymapp subnew) (keymapp subold)
+		      (derived-mode-merge-keymaps subold subnew)))
+	       (setq i (1- i)))))
+      (setq tail (cdr tail))))
   (setcdr (nthcdr (1- (length new)) new) old))
 
 (defun derived-mode-merge-syntax-tables (old new)
   "Merge an old syntax table into a new one.
 Where the new table already has an entry, nothing is copied from the old one."
-  (let ((idx 0)
-	(end (min (length new) (length old))))
-    (while (< idx end)
-      (if (not (aref new idx))
-	  (aset new idx (aref old idx)))
-      (setq idx (1+ idx)))))
+  (set-char-table-parent new old))
 
 ;; Merge an old abbrev table into a new one.
 ;; This function requires internal knowledge of how abbrev tables work,
